# -*- coding: utf-8 -*-
"""
renoir.parsing.stack
--------------------

Provides stack helpers for templates parsing.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

import uuid
from collections import namedtuple
from pathlib import Path

from .contents import Content, Elements, Node, NodeGroup


ParsedLines = namedtuple("ParsedLines", ("start", "end"))


class State:
    __slots__ = [
        "_id",
        "name",
        "source",
        "lines",
        "elements",
        "content",
        "blocks",
        "in_python_block",
        "parent",
        "dependencies",
        "indent",
        "offset",
        "settings",
    ]

    def __init__(
        self, name, elements, in_python_block=False, parent=None, source=None, line_start=1, blocks=None, **settings
    ):
        self._id = uuid.uuid4().hex
        self.name = name
        self.elements = elements
        self.in_python_block = in_python_block
        self.parent = parent
        self.source = source
        self.lines = ParsedLines(line_start, line_start)
        self.settings = settings
        self.content = Content()
        self.blocks = blocks if blocks is not None else {}
        self.dependencies = {}
        self.indent = 0
        self.offset = 0
        if self.elements and self.elements[0].is_python_block and not self.in_python_block:
            self.swap_block_type()

    def __call__(
        self, name=None, elements=None, in_python_block=None, parent=None, source=None, line_start=None, **kwargs
    ):
        name = name or self.name
        elements = self.elements if elements is None else elements
        parent = parent or self
        source = source or parent.source
        settings = dict(**self.settings)
        if in_python_block is None:
            self.swap_block_type()
            in_python_block = parent.in_python_block
            line_start = parent.lines.end if line_start is None else line_start
            settings["isolated_pyblockstate"] = False
        else:
            line_start = 1 if line_start is None else line_start
            settings["isolated_pyblockstate"] = True
        if kwargs:
            settings.update(kwargs)
        return self.__class__(
            name,
            elements,
            in_python_block=in_python_block,
            parent=parent,
            source=source,
            line_start=line_start,
            **settings,
        )

    def swap_block_type(self):
        self.in_python_block = not self.in_python_block

    def update_lines_count(self, additional_lines, offset=None):
        start = self.lines.end if offset is None else offset
        self.lines = self.lines._replace(start=start, end=start + additional_lines)

    def __getattr__(self, name):
        return self.settings.get(name)


class Context:
    def __init__(self, parser, name, text, scope, writer_node_cls, plain_node_cls):
        self.parser = parser
        self.stack = []
        self.scope = scope
        self.state = State(
            name,
            Elements(self.parser._tag_split_text(text)).to_list(),
            source=name,
            isolated_pyblockstate=True,
            new_line=False,
        )
        self.nodes_map = {}
        self._writer_node_cls = writer_node_cls
        self._plain_node_cls = plain_node_cls

    @property
    def name(self):
        return self.state.name

    @property
    def cwd(self):
        return Path(self.state.source).parent

    @property
    def content(self):
        return self.state.content

    @property
    def elements(self):
        return self.state.elements

    def swap_block_type(self):
        return self.state.swap_block_type()

    def update_lines_count(self, *args, **kwargs):
        return self.state.update_lines_count(*args, **kwargs)

    def __call__(self, name=None, elements=None, in_python_block=None, **kwargs):
        self.stack.append(self.state)
        self.state = self.state(name=name, elements=elements, in_python_block=in_python_block, **kwargs)
        return self

    def load(self, name, **kwargs):
        name, file_path, preload_params, text = self.parser._get_file_text(
            self, name, ctxpath=self.cwd, strip_ending_new_line=kwargs.pop("strip_ending_new_line", False)
        )
        self.state.dependencies[name] = preload_params
        kwargs["source"] = file_path
        kwargs["in_python_block"] = False
        return self(name=name, elements=Elements(self.parser._tag_split_text(text)).to_list(), **kwargs)

    def end_current_step(self):
        self.state.elements = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            raise
        self.swap_block_type()
        deps = dict(self.state.dependencies)
        contents = self.content.contents
        name = self.name
        lines = self.state.lines
        in_python_block = self.state.in_python_block
        isolated_pyblockstate = self.state.isolated_pyblockstate
        state_id = self.state._id
        self.state = self.stack.pop()
        node = self.node_group(contents)
        if not isolated_pyblockstate:
            self.state.in_python_block = in_python_block
            self.update_lines_count(lines.end - lines.start, offset=lines.end)
        self.state.blocks[name] = state_id
        self.state.dependencies.update(deps)
        self.nodes_map[state_id] = node

    def python_node(self, value=None):
        node = Node(value, source=self.state.source, lines=self.state.lines)
        self.content.append(node)
        return node

    def variable(self, value=None):
        node = self._writer_node_cls(value, indent=self.state.indent, source=self.state.source, lines=self.state.lines)
        self.content.append(node)
        return node

    def node_group(self, value=None):
        node = NodeGroup(value, indent=self.state.indent)
        self.content.append(node)
        return node

    def _plain(self, node_cls, value, **kwargs):
        self.content.append(
            node_cls(value, indent=self.state.indent, source=self.state.source, lines=self.state.lines, **kwargs)
        )

    def plain(self, value):
        self._plain(self._plain_node_cls, value)

    def parse(self):
        while self.elements:
            element = self.elements.pop(0)
            if self.state.in_python_block:
                self.parser.parse_python_block(self, element)
            else:
                self.parser.parse_plain_block(self, element)
            self.swap_block_type()

    def ignore(self):
        while self.elements:
            element = self.elements.pop(0)
            if self.state.in_python_block:
                self.parser.parse_raw_block(self, element)
            else:
                self.parser.parse_plain_block(self, element)
            self.swap_block_type()


class HTMLContext(Context):
    def __init__(self, parser, name, text, scope, writer_node_cls, plain_node_cls, escape_node_cls):
        super().__init__(parser, name, text, scope, writer_node_cls, plain_node_cls)
        self._escape_node_cls = escape_node_cls

    def variable(self, value=None, escape=True):
        node_cls = self._escape_node_cls if escape else self._writer_node_cls
        node = node_cls(value, indent=self.state.indent, source=self.state.source, lines=self.state.lines)
        self.content.append(node)
        return node

    def html(self, value):
        return self.plain(value)



# -*- coding: utf-8 -*-
"""
renoir.parsing.contents
-----------------------

Provides structures for templating system.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import List, Optional

from .._shortcuts import to_unicode
from ..helpers import adict


class Element:
    __slots__ = [
        "ctx",
        "idx",
        "text",
        "is_python_block",
        "linesn",
        "linesd",
        "strippable_head",
        "strippable_tail",
        "stripped_head",
        "stripped_tail",
        "reindent_skip",
    ]

    def __init__(self, ctx: "Elements", idx: int, text: str, is_python_block: bool = False):
        self.ctx = ctx
        self.idx = idx
        self.text = text
        self.is_python_block = is_python_block
        self.linesn = text.count("\n")
        self.linesd = self.linesn + 1
        self.strippable_head = False
        self.strippable_tail = False
        self.stripped_head = False
        self.stripped_tail = False
        self.reindent_skip = False
        if not text:
            self.strippable_head = False
            self.strippable_tail = False
        else:
            lsplit = text.split("\n", 1)
            rsplit = text.rsplit("\n", 1)
            if not lsplit[0].strip(" "):
                self.strippable_head = True
            if not rsplit[-1].strip(" "):
                self.strippable_tail = True

    def prev(self, positions: int = 1) -> Optional["Element"]:
        idx = self.idx - positions
        if idx < 0:
            return None
        try:
            return self.ctx[idx]
        except IndexError:
            return None

    def next(self, positions: int = 1) -> Optional["Element"]:
        try:
            return self.ctx[self.idx + positions]
        except IndexError:
            return None

    def strip(self, force_reindent_skip: bool = False):
        prev_element, next_element = self.prev(), self.next()
        if prev_element is None and next_element is not None:
            next_element.strip_head()
            next_element.reindent_skip = force_reindent_skip
            return
        if next_element is None:
            return
        if prev_element.strippable_tail and next_element.strippable_head:
            prev_element.strip_tail()
            next_element.strip_head()
            next_element.reindent_skip = force_reindent_skip

    def strip_head(self):
        if not self.strippable_head:
            return
        self.stripped_head = True

    def strip_tail(self):
        if not self.strippable_tail:
            return
        self.stripped_tail = True

    def split(self) -> ElementSplitted:
        return ElementSplitted(self)

    def __str__(self) -> str:
        rv = self.text
        if self.stripped_tail:
            rv = rv.rsplit("\n", 1)[0] + "\n"
        if self.stripped_head:
            rv = rv.split("\n", 1)[-1]
        return rv

    def __bool__(self) -> bool:
        return bool(self.text)

    @property
    def can_arbitrate_reindent(self) -> bool:
        return not bool(str(self))


class ElementSplitted:
    __slots__ = ["parent", "lines"]

    def __init__(self, parent: Element):
        self.parent = parent
        self.lines = [
            adict(text=line, indent=0, original_indent=0, ignore_reindent=False, offset=idx)
            for idx, line in enumerate(self.parent.text.split("\n"))
        ]

    @property
    def linesn(self) -> int:
        return self.parent.linesn

    @property
    def linesd(self) -> int:
        return self.parent.linesd

    def increment_indent(self, increment: int):
        for line in self.lines:
            line.indent += increment

    def change_indent(self, indent: int):
        for line in self.lines:
            line.indent = indent

    def _has_reindent_arbiter(self):
        if self.parent.reindent_skip:
            return False
        if self.parent.stripped_head and self.parent.idx == self.parent.ctx.strip_arbiter.idx:
            return True
        rv, prev = False, self.parent.prev(2)
        while prev is not None:
            if prev.stripped_head and not prev.reindent_skip:
                if prev.idx != self.parent.ctx.strip_arbiter.idx:
                    prev = prev.prev(2)
                else:
                    rv = prev.can_arbitrate_reindent
                    break
            else:
                break
        return rv

    def __str__(self) -> str:
        lines = []
        offsets = (1 if self.parent.stripped_head else None, -1 if self.parent.stripped_tail else None)
        if not self.parent.strippable_head and self.lines:
            self.lines[0].ignore_reindent = True
        if self.parent.stripped_head and self._has_reindent_arbiter():
            self.lines[offsets[0] or 0].ignore_reindent = True
        for line in self.lines[offsets[0] : offsets[1]]:
            lines.append(
                "{pre}{text}".format(
                    pre=(" " * line.original_indent if line.ignore_reindent else " " * line.indent), text=line.text
                )
            )
        if self.parent.stripped_tail:
            lines.append("")
        return "\n".join(lines)

    def __bool__(self) -> bool:
        return bool(str(self))


class Elements(Sequence):
    __slots__ = ["data"]

    def __init__(self, elements: List[str]):
        self.data = []
        in_python_block = False
        offsets = [None, None]
        if len(elements) > 1:
            if not elements[0]:
                offsets[0] = 1
                in_python_block = True
            if not elements[-1]:
                offsets[1] = -1
        for idx, element in enumerate(elements[offsets[0] : offsets[1]]):
            self.data.append(Element(self, idx, element, in_python_block))
            in_python_block = not in_python_block

    @property
    def strip_arbiter(self):
        if not self.data:
            return None
        if not self.data[0].is_python_block:
            return self.data[0]
        if len(self.data) > 1:
            return self.data[1]

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, key: int) -> Element:
        return self.data[key]

    def to_list(self) -> List[Element]:
        return list(self.data)


class Content:
    __slots__ = ["_contents", "_evicted"]

    def __init__(self):
        self._contents = []
        self._evicted = False

    def append(self, element):
        self._contents.append(element)

    def extend(self, *elements):
        for element in elements:
            self.append(element)

    def evict(self):
        self._evicted = True

    @property
    def contents(self):
        return [] if self._evicted else list(self._contents)

    def render(self, parser):
        return "" if self._evicted else "".join(element.__render__(parser) for element in self._contents)

    def reference(self):
        rv = []
        for element in self._contents:
            rv.extend(element.__reference__())
        return rv


class Node:
    __slots__ = ["value", "indent", "source", "lines"]

    def __init__(self, value=None, indent=0, source=None, lines=None):
        self.value = value
        self.indent = indent
        self.source = source
        self.lines = lines or (None, None)

    def increment_indent(self, increment):
        self.indent += increment

    def change_indent(self, indent):
        self.indent = indent

    def __render__(self, parser):
        return "\n" + to_unicode(self.value)

    def __reference__(self):
        return [(self.source, self.lines)]

    def _rendered_lines(self):
        return self.__render__(adict(writer="w")).split("\n")[1:]


class NodeGroup(Node):
    __slots__ = ["value", "indent", "source", "lines", "_evicted"]

    def __init__(self, value=None, **kwargs):
        super().__init__(value or [], **kwargs)
        self._evicted = False

    def increment_children_indent(self, increment):
        for element in self.value:
            element.increment_indent(increment)

    def increment_indent(self, increment):
        self.increment_children_indent(increment)
        self.indent += increment

    def change_indent(self, indent):
        diff = indent - self.indent
        self.increment_children_indent(diff)
        self.indent = indent

    def evict(self):
        self._evicted = True

    def __render__(self, parser):
        return "" if self._evicted else "".join(element.__render__(parser) for element in self.value)

    def __reference__(self):
        rv = []
        stack = self.value if not self._evicted else []
        for element in stack:
            rv.extend(element.__reference__())
        return rv


class WriterNode(Node):
    __slots__ = ["value", "indent", "source", "lines"]
    _writer_method = "write"

    def render_value(self):
        return str(self.value)

    def __render__(self, parser):
        v = to_unicode(self.render_value())
        return f"\n{parser.writer}.{self._writer_method}({v})" if v else ""

    def __reference__(self):
        if not to_unicode(self.render_value()):
            return []
        return super().__reference__()


class PlainNode(WriterNode):
    __slots__ = ["value", "indent", "source", "lines"]

    def render_value(self):
        v = str(self.value)
        return repr(v) if v else None


class WrappedNode(PlainNode):
    __slots__ = ["value", "indent", "source", "lines"]

    def increment_indent(self, increment):
        self.value.increment_indent(increment)
        self.indent += increment

    def change_indent(self, indent):
        diff = indent - self.indent
        self.value.increment_indent(diff)
        self.indent = indent


class HTMLEscapeNode(WriterNode):
    __slots__ = ["value", "indent", "source", "lines"]
    _writer_method = "escape"



# -*- coding: utf-8 -*-
"""
renoir.constants
----------------

Provides constants for the templating system.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

from enum import Enum


NOFILEPATH = "<string>"


class MODES(str, Enum):
    html = "html"
    plain = "plain"


class ESCAPES(str, Enum):
    all = "all"
    common = "common"



# -*- coding: utf-8 -*-
"""
renoir.parsing.parser
---------------------

Provides the templating parser.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

import os
import re
from pathlib import Path

from ..errors import TemplateError
from .contents import HTMLEscapeNode, PlainNode, WrappedNode, WriterNode
from .lexers import default_lexers
from .stack import Context, HTMLContext


class TemplateParser:
    _nodes_cls = {"writer": WriterNode, "plain": PlainNode}

    re_multiline = re.compile(r'(""".*?""")|(\'\'\'.*?\'\'\')', re.DOTALL)

    #: re-indentation rules
    re_auto_dedent = re.compile(r"^(elif |else:|case |except:|except |finally:).*$", re.DOTALL)
    re_dedent = re.compile(r"^(return|continue|break|raise)( .*)?$", re.DOTALL)
    re_pass = re.compile(r"^pass( .*)?$", re.DOTALL)

    def __init__(
        self, templater, text, name="ParserContainer", scope={}, writer="__writer__", lexers={}, delimiters=("{{", "}}")
    ):
        self.templater = templater
        self.name = name
        self.text = text
        self.writer = writer
        self.scope = scope
        #: lexers to use
        self.lexers = dict(default_lexers)
        self.lexers.update(lexers)
        #: configure delimiters
        self.delimiters = delimiters
        escaped_delimiters = (re.escape(delimiters[0]), re.escape(delimiters[1]))
        self.r_tag = re.compile(
            r"((?<!%s)%s.*?%s(?!%s))"
            % (escaped_delimiters[0][0:2], escaped_delimiters[0], escaped_delimiters[1], escaped_delimiters[1][-2:]),
            re.DOTALL,
        )
        self.delimiters_len = (len(self.delimiters[0]), len(self.delimiters[1]))
        #: build content
        self.parse(text)

    def _tag_split_text(self, text):
        return self.r_tag.split(text.replace("\t", "    "))

    def _get_file_text(self, ctx, filename, ctxpath=None, strip_ending_new_line=False):
        #: remove quotation from filename string
        try:
            filename = eval(filename, self.scope)
        except Exception:
            raise TemplateError("Invalid template filename", ctx.state.source, ctx.state.lines)
        #: resolve paths
        preload_name, preload_params = filename, {}
        if any(filename.startswith(relpath) for relpath in ["./", "../"]):
            full_path = (ctxpath / Path(filename)).resolve()
            preload_params["path"] = full_path.parent
            preload_name = full_path.name
        #: get the file contents
        path, file_name = self.templater.preload(preload_name, **preload_params)
        file_path = os.path.join(path, file_name)
        try:
            text = self.templater.load(file_path)
        except Exception:
            raise TemplateError("Unable to open included view file", ctx.state.source, ctx.state.lines)
        text = self.templater.prerender(text, file_path)
        if strip_ending_new_line and text.endswith("\n"):
            text = text[:-1]
        return filename, file_path, (preload_name, preload_params), text

    def parse_plain_block(self, ctx, element):
        ctx.update_lines_count(element.linesn)
        ctx.plain(element)

    #: get rid of delimiters
    def _get_python_block_text(self, element):
        return element.text[self.delimiters_len[0] : -self.delimiters_len[1]].strip()

    #: escape new lines on python comment blocks
    def _escape_python_multiline_newlines(self, text):
        return re.sub(self.re_multiline, _escape_newlines, text)

    def _parse_python_line(self, ctx, element, line):
        #: get line components for lexers
        if line.startswith("="):
            lex, value = "=", line[1:].strip()
        else:
            v = line.split(" ", 1)
            if len(v) == 1:
                lex = v[0]
                value = ""
            else:
                lex = v[0]
                value = v[1]
        #: use appropriate lexer if available for current lex
        lexer = self.lexers.get(lex)
        if lexer and not value.startswith("="):
            if lexer.remove_line:
                element.strip(not lexer.follows_reindent_on_line_removal)
            lexer(ctx, value=value)
            return
        #: otherwise add as a python node
        element.strip()
        ctx.python_node(line)

    def parse_python_block(self, ctx, element):
        text = self._get_python_block_text(element)
        if not text:
            return
        ctx.update_lines_count(element.linesn)
        for line in self._escape_python_multiline_newlines(text).split("\n"):
            self._parse_python_line(ctx, element, line.strip())

    def _parse_raw_contents(self, ctx, element, parsed):
        #: use appropriate lexer
        if parsed == "end":
            lexer = self.lexers[parsed]
            if lexer.remove_line:
                element.strip(not lexer.follows_reindent_on_line_removal)
            lexer(ctx, value=None)
            return
        #: otherwise add as a plain node
        ctx.plain(element)

    def parse_raw_block(self, ctx, element):
        text = self._get_python_block_text(element)
        if not text:
            return
        ctx.update_lines_count(element.linesn)
        self._parse_raw_contents(ctx, element, text.strip())

    def _build_ctx(self, text):
        return Context(
            self,
            self.name,
            text,
            self.scope,
            writer_node_cls=self._nodes_cls["writer"],
            plain_node_cls=self._nodes_cls["plain"],
        )

    def parse(self, text):
        ctx = self._build_ctx(text)
        ctx.parse()
        self.content = ctx.content
        self.dependencies = dict(ctx.state.dependencies)

    def reindent(self, text):
        lines = text.split("\n")
        new_lines = []
        indent = 0
        dedented = 0
        match_stack = []
        #: parse lines
        for raw_line in lines:
            line = raw_line.strip()
            if not line:
                continue
            #: apply auto dedenting
            if self.re_auto_dedent.match(line):
                indent = indent + dedented - 1
                if line.startswith("case") and match_stack:
                    match_stack[-1] -= 1
            dedented = 0
            #: apply indentation
            indent = max(indent, 0)
            new_lines.append(" " * (4 * indent) + line)
            #: dedenting on `pass`
            if self.re_pass.match(line):
                indent -= 1
                if match_stack:
                    match_item = match_stack.pop()
                    if match_item == 0:
                        indent -= 1
            #: implicit dedent on specific commands
            if self.re_dedent.match(line):
                dedented = 1
                indent -= 1
            #: indenting on lines ending with `:`
            if line.endswith(":") and not line.startswith("#"):
                indent += 1
                if line.startswith("match"):
                    match_stack.append(0)
                    indent += 1
                if line.startswith("case") and match_stack:
                    match_stack[-1] += 1
        #: handle indentation errors
        if indent > 0:
            raise TemplateError('missing "pass" in view', self.name, 1)
        elif indent < 0:
            raise TemplateError('too many "pass" in view', self.name, 1)
        #: rebuild text
        return "\n".join(new_lines)

    def render(self):
        return self.reindent(self.content.render(self))


class IndentTemplateParser(TemplateParser):
    re_wspace = re.compile("^( *)")

    def parse_plain_block(self, ctx, element):
        current_line, new_line = ctx.state.lines.end, False
        lines_element = element.split()
        ctx.update_lines_count(lines_element.linesn)
        if ctx.state.lines.start != current_line:
            new_line = True
        for line in lines_element.lines:
            if line.offset:
                new_line = True
            if not new_line:
                continue
            indent = len(self.re_wspace.search(line.text).group(0))
            ctx.state.indent = indent
            ctx.state.offset = len(line.text) - indent
            line.text = line.text[indent:]
            line.indent = ctx.state.indent
            new_line = False
        ctx._plain(WrappedNode, lines_element)


class HTMLTemplateParser(TemplateParser):
    _nodes_cls = {"writer": WriterNode, "plain": PlainNode, "escape": HTMLEscapeNode}

    def _build_ctx(self, text):
        return HTMLContext(
            self,
            self.name,
            text,
            self.scope,
            writer_node_cls=self._nodes_cls["writer"],
            plain_node_cls=self._nodes_cls["plain"],
            escape_node_cls=self._nodes_cls["escape"],
        )


class HTMLIndentTemplateParser(HTMLTemplateParser, IndentTemplateParser):
    @staticmethod
    def _html_pre_limiters(ctx, line):
        if not ctx.state.in_html_pre and "<pre" in line and "</pre>" not in line:
            return True, False
        if ctx.state.in_html_pre and "</pre>" in line:
            return False, True
        return False, False

    @staticmethod
    def _start_html_pre(ctx, start):
        if start:
            ctx.state.settings["in_html_pre"] = True

    @staticmethod
    def _end_html_pre(ctx, end):
        if end:
            ctx.state.settings["in_html_pre"] = False

    def parse_plain_block(self, ctx, element):
        lines_element = element.split()
        ctx.update_lines_count(lines_element.linesn)
        for line in lines_element.lines:
            indent = len(self.re_wspace.search(line.text).group(0))
            ctx.state.indent = indent
            start_pre, end_pre = self._html_pre_limiters(ctx, line.text)
            self._end_html_pre(ctx, end_pre)
            if ctx.state.in_html_pre:
                line.text = line.text[indent:]
                line.original_indent = ctx.state.indent
                line.ignore_reindent = True
            self._start_html_pre(ctx, start_pre)
        ctx._plain(WrappedNode, lines_element)


def _escape_newlines(re_val):
    #: take the entire match and replace newlines with escaped newlines
    return re_val.group(0).replace("\n", "\\n")



# -*- coding: utf-8 -*-
"""
renoir.parsing.lexers
---------------------

Provides lexers for templates parsing.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

from typing import Optional

from .stack import Context


class Lexer:
    evaluate: bool = False
    remove_line: bool = False
    follows_reindent_on_line_removal: bool = True

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __call__(self, ctx: Context, value: Optional[str] = None):
        if self.evaluate and value is not None:
            value = eval(value, ctx.scope)
        self.process(ctx, value)

    def process(self, ctx: Context, value: Optional[str]):
        raise NotImplementedError


class VariableLexer(Lexer):
    def process(self, ctx, value):
        #: insert a variable in the template
        ctx.variable(value)


class BlockLexer(Lexer):
    remove_line = True
    follows_reindent_on_line_removal = False

    def process(self, ctx, value):
        #: create a new stack element with name
        with ctx(value):
            ctx.parse()


class EndLexer(Lexer):
    remove_line = True

    def process(self, ctx, value):
        #: we are done with this node, move up in the stack
        ctx.end_current_step()


class SuperLexer(Lexer):
    remove_line = True

    def process(self, ctx, value):
        #: create a node for later injection by super block
        target_block = value if value else ctx.name
        node = ctx.node_group()
        ctx.state.injections[ctx.state.extend_src_id][target_block] = node


class IncludeLexer(Lexer):
    def process(self, ctx, value):
        #: if we have a value, just add the new content
        if value:
            with ctx.load(value, strip_ending_new_line=True):
                ctx.parse()
                included_id = ctx.state._id
        #: otherwise, inject in the extended node
        else:
            extend_src = ctx.state.extend_map[ctx.state.source]
            if extend_src._id in ctx.state.includes_parsed:
                node = ctx.node_group(ctx.state.includes_parsed[extend_src._id].value)
                node.increment_children_indent(ctx.state.indent + ctx.state.offset)
                return
            extend_src.swap_block_type()
            with ctx(
                f"__include__{extend_src._id}",
                extend_src.elements,
                in_python_block=extend_src.in_python_block,
                source=extend_src.source,
                line_start=extend_src.lines.end,
                blocks=extend_src.blocks,
                extend_src_id=extend_src._id,
            ):
                ctx.parse()
                ctx.state.blocks_map[extend_src._id].update(ctx.state.blocks)
                extend_src.update_lines_count(ctx.state.lines.end - ctx.state.lines.start)
                included_id = ctx.state._id
            ctx.state.implicit_extenders.pop(extend_src._id)
            ctx.state.includes_parsed[extend_src._id] = ctx.nodes_map[included_id]
        ctx.nodes_map[included_id].increment_children_indent(ctx.state.indent + ctx.state.offset)


class ExtendLexer(Lexer):
    remove_line = True

    def process(self, ctx, value):
        #: extend the proper template
        with ctx.load(
            value,
            blocks_map=ctx.state.blocks_map or {},
            extend_map=ctx.state.extend_map or {},
            implicit_extenders=ctx.state.implicit_extenders or {},
            includes_parsed=ctx.state.includes_parsed or {},
            injections=ctx.state.injections or {},
        ):
            ctx.state.blocks_map[ctx.state.parent._id] = {}
            ctx.state.extend_map[ctx.state.source] = ctx.state.parent
            ctx.state.implicit_extenders[ctx.state.parent._id] = True
            ctx.state.injections[ctx.state.parent._id] = {}
            ctx.parse()
            if ctx.state.implicit_extenders.pop(ctx.state.parent._id, None):
                self._parse_implicit_extender(ctx)
            self.inject_content_in_children(ctx, ctx.state.injections[ctx.state.parent._id])
            self.replace_extended_blocks(ctx, ctx.state.blocks_map[ctx.state.parent._id])
            ctx.state.injections.pop(ctx.state.parent._id)

    def _parse_implicit_extender(self, ctx):
        extend_src = ctx.state.extend_map[ctx.state.source]
        extend_src.swap_block_type()
        with ctx(
            f"__impl__{extend_src._id}",
            extend_src.elements,
            in_python_block=extend_src.in_python_block,
            source=extend_src.source,
            line_start=extend_src.lines.end,
            blocks=extend_src.blocks,
            extend_src_id=extend_src._id,
        ):
            ctx.parse()
            ctx.state.blocks_map[extend_src._id].update(ctx.state.blocks)
            ctx.content.evict()
            extend_src.update_lines_count(ctx.state.lines.end - ctx.state.lines.start)

    def inject_content_in_children(self, ctx, injections):
        for key, node in injections.items():
            #: get the content to inject
            src = ctx.nodes_map[ctx.state.blocks[key]]
            original_indent = src.indent
            #: align src indent with the destination
            src.change_indent(node.indent)
            node.value = list(src.value)
            #: restore the original indent on the block
            src.indent = original_indent

    def replace_extended_blocks(self, ctx, blocks_map):
        for key in set(ctx.state.blocks.keys()) & set(blocks_map.keys()):
            #: get destination and source blocks
            dst = ctx.state.blocks[key]
            src = blocks_map[key]
            #: update the source indent with the destination one
            ctx.nodes_map[src].change_indent(ctx.nodes_map[dst].indent)
            ctx.nodes_map[dst].value = list(ctx.nodes_map[src].value)
            #: cleanup
            ctx.nodes_map[src].evict()
            ctx.nodes_map[src] = ctx.nodes_map[dst]


class IgnoreLexer(Lexer):
    remove_line = True
    follows_reindent_on_line_removal = False

    def process(self, ctx, value):
        with ctx("__ignore__"):
            ctx.ignore()


default_lexers = {
    "=": VariableLexer(),
    "block": BlockLexer(),
    "end": EndLexer(),
    "super": SuperLexer(),
    "include": IncludeLexer(),
    "extend": ExtendLexer(),
    "raw": IgnoreLexer(),
}



# -*- coding: utf-8 -*-
"""
tests.cache
-----------

Tests cache module.
"""

import pytest

from renoir import Renoir


@pytest.fixture(scope="function")
def templater_reload():
    return Renoir(reload=True)


@pytest.fixture(scope="function")
def templater_noreload():
    return Renoir()


def test_noreload(templater_noreload):
    assert not templater_noreload.cache.parse.data

    templater_noreload._render(source="{{=a}}", context={"a": 1})
    assert templater_noreload.cache.parse.data["<string>"]
    assert not templater_noreload.cache.parse.hashes
    data = templater_noreload.cache.parse.data["<string>"]

    templater_noreload._render(source="{{=a}}", context={"a": 1})
    assert templater_noreload.cache.parse.data["<string>"] is data

    templater_noreload._render(source="{{=a}}\n", context={"a": 1})
    assert templater_noreload.cache.parse.data["<string>"] is data


def test_reload(templater_reload):
    assert not templater_reload.cache.parse.data

    templater_reload._render(source="{{=a}}", context={"a": 1})
    assert templater_reload.cache.parse.data["<string>"]
    data = templater_reload.cache.parse.data["<string>"]
    hashed = templater_reload.cache.parse.hashes["<string>"]

    templater_reload._render(source="{{=a}}", context={"a": 1})
    assert templater_reload.cache.parse.hashes["<string>"] == hashed
    assert templater_reload.cache.parse.data["<string>"] is data

    templater_reload._render(source="{{=a}}\n", context={"a": 1})
    assert templater_reload.cache.parse.hashes["<string>"] != hashed
    assert templater_reload.cache.parse.data["<string>"] is not data



# -*- coding: utf-8 -*-
"""
renoir.writers
--------------

Provides the writers for the templating system.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

from io import StringIO

from ._shortcuts import htmlescape, to_bytes, to_unicode


class Writer:
    def __init__(self):
        self.body = StringIO()

    @staticmethod
    def _to_html(data):
        return htmlescape(data)

    @staticmethod
    def _to_unicode(data):
        if isinstance(data, str):
            return data
        return to_unicode(data)

    def write(self, data):
        self.body.write(self._to_unicode(data))

    def _escape_data(self, data):
        body = None
        if hasattr(data, "__html__"):
            try:
                body = data.__html__()
            except Exception:
                pass
        if body is None:
            body = self._to_html(self._to_unicode(data))
        return body

    def escape(self, data):
        self.write(self._escape_data(data))


class EscapeAll:
    @staticmethod
    def _to_html(data):
        return to_bytes(Writer._to_html(data), "ascii", "xmlcharrefreplace")


class EscapeAllWriter(EscapeAll, Writer):
    pass



# -*- coding: utf-8 -*-
"""
renoir.errors
-------------

Provides errors.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

from .helpers import TemplateReference


class TemplateMissingError(Exception):
    def __init__(self, file_path):
        self.path = file_path
        self.message = f"Template {self.path} not found"
        super().__init__(self.message)


class TemplateError(Exception):
    def __init__(self, message, file_path, lineno):
        super().__init__(message)
        self.message = message
        self.file_path = file_path
        if isinstance(lineno, tuple):
            lineno = lineno[0]
        self.lineno = lineno


class TemplateSyntaxError(Exception):
    def __init__(self, parser_ctx, exc_type, exc_value, tb):
        super().__init__("invalid syntax")
        self._reference = TemplateReference(parser_ctx, exc_type, exc_value, tb)

    @property
    def file_path(self):
        return self._reference.file_path

    @property
    def lineno(self):
        return self._reference.lineno

    @property
    def message(self):
        location = f'File "{self.file_path}", line {self.lineno}'
        lines = [self.args[0], "  " + location]
        return "\n".join(lines)



# -*- coding: utf-8 -*-
"""
renoir._shortcuts
-----------------

Some shortcuts

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

import hashlib
import html


hashlib_sha1 = lambda s: hashlib.sha1(bytes(s, "utf8"))


def to_bytes(obj, charset="utf8", errors="strict"):
    if obj is None:
        return None
    if isinstance(obj, (bytes, bytearray, memoryview)):
        return bytes(obj)
    if isinstance(obj, str):
        return obj.encode(charset, errors)
    raise TypeError("Expected bytes")


def to_unicode(obj, charset="utf8", errors="strict"):
    if obj is None:
        return None
    if not isinstance(obj, bytes):
        return str(obj)
    return obj.decode(charset, errors)


def to_str(obj):
    if not isinstance(obj, str):
        return str(obj)
    return obj


def htmlescape(obj):
    if hasattr(obj, "__html__"):
        return obj.__html__()
    return html.escape(to_str(obj), True).replace("'", "&#39;")



# -*- coding: utf-8 -*-
"""
renoir.helpers
--------------

Provides helpers.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

import traceback


class TemplateReference:
    def __init__(self, parser_ctx, exc_type, exc_value, tb):
        self.parser_ctx = parser_ctx
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.tb = tb
        if hasattr(exc_value, "lineno"):
            writer_lineno = exc_value.lineno
        else:
            template_frame = traceback.extract_tb(tb, 2)[-1]
            writer_lineno = template_frame[1]
        self.lines = parser_ctx.content.reference()
        self.file_path, self.lineno = self.match_template(writer_lineno)

    def match_template(self, writer_lineno):
        try:
            element = self.lines[writer_lineno - 1]
            reference = (element[0], element[1])
        except Exception:
            reference = (self.parser_ctx.name, ("<unknown>", "unknown"))
        return reference[0], reference[1][0]


class ParserCtx:
    __slots__ = ("name", "content")

    def __init__(self, name, content):
        self.name = name
        self.content = content


class adict(dict):
    __setattr__ = dict.__setitem__
    __getattr__ = dict.__getitem__



# -*- coding: utf-8 -*-
"""
tests.parser
------------

Tests parser module.
"""

import sys
import traceback

import pytest

from renoir import Renoir


@pytest.fixture(scope="function")
def templater_plain():
    return Renoir(mode="plain", debug=True)


@pytest.fixture(scope="function")
def templater_escape():
    return Renoir(escape="all", debug=True)


@pytest.fixture(scope="function", params=[False, True], ids=["basic", "auto-indent"])
def ptemplater_plain(request):
    return Renoir(mode="plain", adjust_indent=request.param, debug=True)


def test_variable(templater_plain):
    assert templater_plain._render(source="{{=1}}") == "1"
    assert templater_plain._render(source="{{=a}}", context={"a": list(range(0, 5))}) == "[0, 1, 2, 3, 4]"


def test_variable_encoding(templater_plain):
    assert (
        templater_plain._render(source="{{=a}}", context={"a": "nuvolosità variabile".encode("utf8")})
        == "nuvolosità variabile"
    )
    assert templater_plain._render(source="{{=a}}", context={"a": "nuvolosità variabile"}) == "nuvolosità variabile"


def test_variable_escaping(templater_escape):
    assert (
        templater_escape._render(source="{{=a}}", context={"a": "nuvolosità variabile".encode("utf8")})
        == "nuvolosit&#224; variabile"
    )
    assert (
        templater_escape._render(source="{{=a}}", context={"a": "nuvolosità variabile"}) == "nuvolosit&#224; variabile"
    )


def test_raw(ptemplater_plain):
    s = "{{=1}}{{raw}}{{=1}}{{end}}"
    assert ptemplater_plain._render(source=s) == "1{{=1}}"


def test_pycode_if(ptemplater_plain):
    s = "{{if a == 1:}}\nfoo\n{{elif a == 2:}}\nbar" "\n{{else:}}\nfoobar\n{{pass}}"
    r = ptemplater_plain._render(source=s, context={"a": 1})
    assert r == "foo\n"
    r = ptemplater_plain._render(source=s, context={"a": 2})
    assert r == "bar\n"
    r = ptemplater_plain._render(source=s, context={"a": 25})
    assert r == "foobar\n"


def test_pycode_for(ptemplater_plain):
    s = "{{for i in range(0, 5):}}\n{{=i}}\n{{pass}}"
    r = ptemplater_plain._render(source=s)
    assert r == "\n".join([str(i) for i in range(0, 5)]) + "\n"


@pytest.mark.skipif(sys.version_info < (3, 10), reason="match requires py >= 3.10")
def test_pycode_match(ptemplater_plain):
    s = (
        "{{match a:}}\n"
        "{{case 1:}}\na cat\n"
        "{{case a if a < 5:}}\nfew cats\n"
        "{{case _:}}\nseveral cats\n"
        "{{pass}}showed up again"
    )
    r = ptemplater_plain._render(source=s, context={"a": 1})
    assert r == "a cat\nshowed up again"
    r = ptemplater_plain._render(source=s, context={"a": 3})
    assert r == "few cats\nshowed up again"
    r = ptemplater_plain._render(source=s, context={"a": 25})
    assert r == "several cats\nshowed up again"


def test_python_errors(templater_plain):
    with pytest.raises(ZeroDivisionError) as exc:
        templater_plain._render("foo\n{{=1/0}}\nbar")

    tbs = "".join(traceback.format_exception(exc.type, exc.value, exc.tb))
    frame = exc.traceback[-1]

    frames = []
    tb = exc.tb
    while tb:
        frames.append(tb)
        tb = tb.tb_next
    tb_frame = frames[-1]

    assert frame.name == "template"
    assert str(frame.path) == "<string>"
    assert tb_frame.tb_lineno == 2

    assert '<string>", line 2, in template' in tbs



# -*- coding: utf-8 -*-
"""
renoir.debug
------------

Provides debugging utilities.

:copyright: 2014 Giovanni Barillari

Quite a lot of magic comes from Flask (http://flask.pocoo.org)
:copyright: (c) 2014 by Armin Ronacher.

:license: BSD-3-Clause
"""

import sys
import traceback
from types import CodeType, TracebackType

from ._internal import reraise
from .errors import TemplateError, TemplateSyntaxError


# on pypy we can take advantage of transparent proxies
try:
    from __pypy__ import tproxy
except ImportError:
    tproxy = None


# how does the raise helper look like?
try:
    exec("raise TypeError, 'foo'")
except SyntaxError:
    raise_helper = "raise __renoir_exception__[1]"
except TypeError:
    raise_helper = "raise __renoir_exception__[0], __renoir_exception__[1]"


class TracebackFrameProxy:
    """Proxies a traceback frame."""

    def __init__(self, tb):
        self.tb = tb
        self._tb_next = None

    @property
    def tb_next(self):
        return self._tb_next

    def set_next(self, next):
        if tb_set_next is not None:
            try:
                tb_set_next(self.tb, next and next.tb or None)
            except Exception:
                # this function can fail due to all the hackery it does
                # on various python implementations.  We just catch errors
                # down and ignore them if necessary.
                pass
        self._tb_next = next

    def __getattr__(self, name):
        return getattr(self.tb, name)


class ProcessedTraceback:
    """Holds a renoir preprocessed traceback for printing or reraising."""

    def __init__(self, exc_type, exc_value, frames):
        assert frames, "no frames for this traceback?"
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.frames = frames

        # newly concatenate the frames (which are proxies)
        prev_tb = None
        for tb in self.frames:
            if prev_tb is not None:
                prev_tb.set_next(tb)
            prev_tb = tb
        prev_tb.set_next(None)

    def render_as_text(self, limit=None):
        """Return a string with the traceback."""
        lines = traceback.format_exception(self.exc_type, self.exc_value, self.frames[0], limit=limit)
        return "".join(lines).rstrip()

    @property
    def exc_info(self):
        """Exception info tuple with a proxy around the frame objects."""
        return self.exc_type, self.exc_value, self.frames[0]

    @property
    def standard_exc_info(self):
        """Standard python exc_info for re-raising"""
        tb = self.frames[0]
        # the frame will be an actual traceback (or transparent proxy) if
        # we are on pypy or a python implementation with support for tproxy
        if type(tb) is not TracebackType:
            tb = tb.tb
        return self.exc_type, self.exc_value, tb


def make_traceback(exc_info):
    initial_skip = 1
    if isinstance(exc_info[1], TemplateError):
        exc_info = translate_template_error(exc_info[1])
        initial_skip = 0
    elif isinstance(exc_info[1], TemplateSyntaxError):
        exc_info = translate_syntax_error(exc_info[1])
        initial_skip = 0
    tb = translate_exception(exc_info, initial_skip)
    exc_type, exc_value, tb = tb.standard_exc_info
    reraise(exc_type, exc_value, tb)


def translate_syntax_error(error):
    """Rewrites a syntax error to please traceback systems."""
    exc_info = (error.__class__, error, None)
    filename = error.file_path
    if filename is None:
        filename = "<unknown>"
    return fake_exc_info(exc_info, filename, error.lineno)


def translate_template_error(error):
    exc_info = (error.__class__, error, None)
    filename = error.file_path
    if filename is None:
        filename = "<unknown>"
    return fake_exc_info(exc_info, filename, error.lineno)


def make_frame_proxy(frame):
    proxy = TracebackFrameProxy(frame)
    if tproxy is None:
        return proxy

    def operation_handler(operation, *args, **kwargs):
        if operation in ("__getattribute__", "__getattr__"):
            return getattr(proxy, args[0])
        elif operation == "__setattr__":
            proxy.__setattr__(*args, **kwargs)
        else:
            return getattr(proxy, operation)(*args, **kwargs)

    return tproxy(TracebackType, operation_handler)


def translate_exception(exc_info, initial_skip=0):
    """If passed an exc_info it will automatically rewrite the exceptions
    all the way down to the correct line numbers and frames.
    """
    tb = exc_info[2]
    frames = []

    is_template_error = isinstance(exc_info[1], (TemplateError, TemplateSyntaxError))

    # skip some internal frames if wanted
    for _ in range(initial_skip):
        if tb is not None:
            tb = tb.tb_next

    while tb is not None:
        # save a reference to the next frame if we override the current
        # one with a faked one.
        next = tb.tb_next

        # fake template exceptions
        template = tb.tb_frame.f_globals.get("__renoir_template__")
        if template is not None:
            lineno = template.lineno
            tb = fake_exc_info(exc_info[:2] + (tb,), template.file_path, lineno)[2]

        frames.append(make_frame_proxy(tb))
        if is_template_error and "__renoir_template__" in tb.tb_frame.f_globals:
            break
        tb = next

    # if we don't have any exceptions in the frames left, we have to
    # reraise it unchanged.
    # XXX: can we backup here?  when could this happen?
    if not frames:
        reraise(exc_info[0], exc_info[1], exc_info[2])

    return ProcessedTraceback(exc_info[0], exc_info[1], frames)


def fake_exc_info(exc_info, filename, lineno):
    """Helper for `translate_exception`."""
    exc_type, exc_value, tb = exc_info

    # figure the real context out
    if tb is not None:
        real_locals = tb.tb_frame.f_locals.copy()
        ctx = real_locals.get("context")
        if ctx:
            locals = ctx
        else:
            locals = {}
        # for name, value in real_locals.items():
        #    if name.startswith('l_') and value is not missing:
        #        locals[name[2:]] = value

        # if there is a local called __renoir_exception__, we get
        # rid of it to not break the debug functionality.
        locals.pop("__renoir_exception__", None)
    else:
        locals = {}

    # assamble fake globals we need
    globals = {
        "__name__": filename,
        "__file__": filename,
        "__renoir_exception__": exc_info[:2],
        # we don't want to keep the reference to the template around
        # to not cause circular dependencies, but we mark it as renoir
        # frame for the ProcessedTraceback
        "__renoir_template__": None,
    }

    # and fake the exception
    code = compile("\n" * (lineno - 1) + raise_helper, filename, "exec")
    if hasattr(code, "replace"):
        code = code.replace(co_filename=filename, co_name="template")
    elif hasattr(code, "co_kwonlyargcount"):
        code = CodeType(
            0,
            0,
            code.co_nlocals,
            code.co_stacksize,
            code.co_flags,
            code.co_code,
            code.co_consts,
            code.co_names,
            code.co_varnames,
            filename,
            "template",
            code.co_firstlineno,
            code.co_lnotab,
            (),
            (),
        )
    else:
        code = CodeType(
            0,
            code.co_nlocals,
            code.co_stacksize,
            code.co_flags,
            code.co_code,
            code.co_consts,
            code.co_names,
            code.co_varnames,
            filename,
            "template",
            code.co_firstlineno,
            code.co_lnotab,
            (),
            (),
        )

    # execute the code and catch the new traceback
    try:
        exec(code, globals, locals)
    except Exception:
        exc_info = sys.exc_info()
        new_tb = exc_info[2].tb_next

    # return without this frame
    return exc_info[:2] + (new_tb,)


def _init_ugly_crap():
    """This function implements a few ugly things so that we can patch the
    traceback objects.  The function returned allows resetting `tb_next` on
    any python traceback object.  Do not attempt to use this on non cpython
    interpreters
    """
    import ctypes

    # figure out side of _Py_ssize_t
    if hasattr(ctypes.pythonapi, "Py_InitModule4_64"):
        _Py_ssize_t = ctypes.c_int64
    else:
        _Py_ssize_t = ctypes.c_int

    # regular python
    class _PyObject(ctypes.Structure):
        pass

    _PyObject._fields_ = [("ob_refcnt", _Py_ssize_t), ("ob_type", ctypes.POINTER(_PyObject))]

    # python with trace
    if hasattr(sys, "getobjects"):

        class _PyObject(ctypes.Structure):
            pass

        _PyObject._fields_ = [
            ("_ob_next", ctypes.POINTER(_PyObject)),
            ("_ob_prev", ctypes.POINTER(_PyObject)),
            ("ob_refcnt", _Py_ssize_t),
            ("ob_type", ctypes.POINTER(_PyObject)),
        ]

    class _Traceback(_PyObject):
        pass

    _Traceback._fields_ = [
        ("tb_next", ctypes.POINTER(_Traceback)),
        ("tb_frame", ctypes.POINTER(_PyObject)),
        ("tb_lasti", ctypes.c_int),
        ("tb_lineno", ctypes.c_int),
    ]

    def tb_set_next(tb, next):
        """Set the tb_next attribute of a traceback object."""
        if not (isinstance(tb, TracebackType) and (next is None or isinstance(next, TracebackType))):
            raise TypeError("tb_set_next arguments must be traceback objects")
        obj = _Traceback.from_address(id(tb))
        if tb.tb_next is not None:
            old = _Traceback.from_address(id(tb.tb_next))
            old.ob_refcnt -= 1
        if next is None:
            obj.tb_next = ctypes.POINTER(_Traceback)()
        else:
            next = _Traceback.from_address(id(next))
            next.ob_refcnt += 1
            obj.tb_next = ctypes.pointer(next)

    return tb_set_next


# try to get a tb_set_next implementation if we don't have transparent
# proxies.
tb_set_next = None
if tproxy is None:
    try:
        tb_set_next = _init_ugly_crap()
    except Exception:
        pass
    del _init_ugly_crap



# -*- coding: utf-8 -*-
"""
tests.templater
---------------

Tests templater module.
"""

import os
import traceback

import pytest
import yaml

from renoir import Renoir


@pytest.fixture(scope="function")
def templater():
    return Renoir(mode="plain", debug=True, path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "yaml"))


@pytest.fixture(scope="function")
def templater_indent():
    return Renoir(
        mode="plain",
        adjust_indent=True,
        debug=True,
        path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "yaml"),
    )


@pytest.fixture(scope="function")
def templater_html():
    return Renoir(debug=True, path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "html"))


@pytest.fixture(scope="function")
def templater_html_indent():
    return Renoir(
        escape="all",
        adjust_indent=True,
        debug=True,
        path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "html"),
    )


yaml_rendered = """
str: str
int: 1
float: 1.2
sfloat: "1.2"
obj:
  nested:
    key: str
    array:
      - foo
      - bar
  array:
    - foo
    - foo: bar
      array:
        - str
        - foo
        - 1
        - 1.2
    - bar: foo bar
    - bar2: foo foo2 bar
    - bar3: foo foo2 bar bar2
    - bar3: foo foo2 barbar2
    - bar4: foo foo2 barbar2 bar3
"""


def test_plain(templater):
    r = templater.render("basic.yaml", {})
    assert r == yaml_rendered[1:]
    data = yaml.load(r, Loader=yaml.SafeLoader)
    assert data["obj"]["nested"]["array"][1] == "bar"
    assert data["obj"]["array"][1]["array"][-1] == 1.2


yaml_indent_rendered = """
str: str
int: 1
float: 1.2
sfloat: "1.2"
obj:
  nested:
    key: str
    array:
      - foo
      - bar
  array:
    - foo
    - foo: bar
      array: []
    - bar: foo bar
    - bar2: foo foo2 bar
    - bar3: foo foo2 bar bar2
    - bar3: foo foo2 barbar2
    - bar4: foo foo2 barbar2 bar3
inclusion:
  - inclusion_key: inclusion_val
    inclusion_nest:
      first: foo
      second: bar
    main_key: main_val
  - inclusion_1: foo
    main_key: main_val
  - inclusion_2:
      1st: foo
      2nd: bar
    main_key: main_val
  - inclusion_3:
      1st: foo
    main_key: main_val
  - inclusion_4:
      1st: foo
      2nd: bar
    main_key: main_val
added:
  indent:
    foo:
      bar:
        - baz
  ints:
    - 2
    - 4
    - 6
    - 8
"""


def test_plain_indent(templater_indent):
    r = templater_indent.render(
        "nested.yaml",
        {
            "indent": lambda v, i: (
                "\n".join([v.split("\n")[0], "\n".join([f"{' '*i}{el}" for el in v.split("\n")[1:]])])
            ),
            "additional": "foo:\n  bar:\n    - baz",
        },
    )
    assert "\n".join([l.rstrip() for l in r.splitlines()]) == yaml_indent_rendered[1:]
    data = yaml.load(r, Loader=yaml.SafeLoader)
    assert data["obj"]["nested"]["array"][1] == "bar"
    assert data["inclusion"][0]["main_key"] == "main_val"
    assert data["inclusion"][0]["inclusion_nest"]["first"] == "foo"
    assert not data["obj"]["array"][1]["array"]
    assert data["added"]["indent"]["foo"]["bar"][0] == "baz"
    assert data["added"]["ints"][-1] == 8


html_rendered = """
<!DOCTYPE html>
<html>
    <head>
        <title>Test</title>

    </head>
    <body>
        <div>header1</div>
<div>header2</div>
        <div class="page">
            <a href="/" class="title"><h1>Test</h1></a>
            <div class="nav">
                <a href="/">nuvolosità variabile</a>
            </div>


<ul class="posts">
    <li>
        <h2>foo</h2>
        <hr />
    </li>
    <li>
        <h2>bar</h2>
        <hr />
    </li>
</ul>

        </div>
        <div>footer</div>
    </body>
</html>"""


def test_html(templater_html):
    r = templater_html.render("test.html", {"posts": [{"title": "foo"}, {"title": "bar"}]})
    assert "\n".join([l.rstrip() for l in r.splitlines()]) == html_rendered[1:]


html_indent_rendered = """
<!DOCTYPE html>
<html>
    <head>
        <title>Test</title>

    </head>
    <body>
        <div>header1</div>
        <div>header2</div>
        <div class="page">
            <a href="/" class="title"><h1>Test</h1></a>
            <div class="nav">
                <a href="/">nuvolosit&#224; variabile</a>
            </div>


            <ul class="posts">
                <li>
                    <h2>foo</h2>
                    <hr />
                </li>
                <li>
                    <h2>bar</h2>
                    <hr />
                </li>
            </ul>

        </div>
        <div>footer</div>
    </body>
</html>"""


def test_html_indent(templater_html_indent):
    r = templater_html_indent.render("test.html", {"posts": [{"title": "foo"}, {"title": "bar"}]})
    assert "\n".join([l.rstrip() for l in r.splitlines()]) == html_indent_rendered[1:]


html_indent_rendered2 = """
<!DOCTYPE html>
<html>
    <head>
        <title>Test</title>

    </head>
    <body>
        <div>header1</div>
        <div class="page">
            <a href="/" class="title"><h1>Test</h1></a>
            <div class="nav">
                <a href="/">nuvolosit&#224; variabile</a>
            </div>

            <ul class="posts">
                <li>
                    <h2>foo</h2>
                    <hr />
                </li>
                <li>
                    <h2>bar</h2>
                    <hr />
                </li>
            </ul>

        </div>
        <div>footer</div>
    </body>
</html>"""


def test_html_indent2(templater_html_indent):
    r = templater_html_indent.render("test2.html", {"posts": [{"title": "foo"}, {"title": "bar"}]})
    assert "\n".join([l.rstrip() for l in r.splitlines()]) == html_indent_rendered2[1:]


html_pre_rendered = """
<div>
    <pre>
        <code>
var foo = 'foo';
var bar = '{{ bar }}';
var baz = 'baz';
var list = [
    '0',
    '1',
];
        </code>
    </pre>
</div>"""


def test_html_indent_pre(templater_html_indent):
    r = templater_html_indent.render("pre.html")
    assert "\n".join([l.rstrip() for l in r.splitlines()]) == html_pre_rendered[1:]


def test_pyerror(templater_html):
    with pytest.raises(ZeroDivisionError) as exc:
        templater_html.render("pyerror.html")

    tbs = "".join(traceback.format_exception(exc.type, exc.value, exc.tb))
    frame = exc.traceback[-1]

    frames = []
    tb = exc.tb
    while tb:
        frames.append(tb)
        tb = tb.tb_next
    tb_frame = frames[-1]

    tpath = os.sep.join(["html", "pyerror.html"])

    assert frame.name == "template"
    assert str(frame.path).endswith(tpath)
    assert tb_frame.tb_lineno == 2

    assert f'{tpath}", line 2, in template' in tbs


@pytest.fixture(scope="function")
def templater_blocks():
    return Renoir(mode="plain", debug=True, path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "blocks"))


_target_b1 = """
super b2
parent b2
child b2
parent l1
child b1
parent l2
child l3
parent l4
child l1
child l2"""


def test_blocks_include(templater_blocks):
    r = templater_blocks.render("child.txt", {"parent_name": "./parent_incl.txt"})
    assert "\n".join(filter(None, [l.rstrip() for l in r.splitlines()])) == _target_b1[1:]


_target_b2 = """
super b2
parent b2
child b2
parent l1
child b1
parent l2
child l3
parent l4"""


def test_blocks_noinclude(templater_blocks):
    r = templater_blocks.render("child.txt", {"parent_name": "./parent_noincl.txt"})
    assert "\n".join(filter(None, [l.rstrip() for l in r.splitlines()])) == _target_b2[1:]


_target_b3 = """
parent l1
child b1
parent b2
parent c1
child l1
parent l2"""


def test_blocks_include_multi(templater_blocks):
    r = templater_blocks.render("child_multi_incl.txt", {"condition": True})
    assert "\n".join(filter(None, [l.rstrip() for l in r.splitlines()])) == _target_b3[1:]



# -*- coding: utf-8 -*-
"""
renoir.typing
-------------

Provides typing helpers.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

from typing import Any, Callable, Dict, Tuple


LoaderType = Callable[[str, str], Tuple[str, str]]
RenderType = Callable[[str, str], str]
ContextType = Callable[[Dict[str, Any]], None]



# -*- coding: utf-8 -*-
"""
tests.extensions
----------------

Tests extensions module.
"""

import pytest

from renoir import Extension, Lexer, Renoir


class FooLexer(Lexer):
    def process(self, ctx, value):
        ctx.html(value * 2)


class FooExtension(Extension):
    namespace = "foo"
    lexers = {"foo": FooLexer}

    def render(self, source, name):
        return source + "foo"


class BarExtension(Extension):
    namespace = "bar"
    file_extension = "test"

    def load(self, path, file_name):
        return path, file_name + ".bar"

    def context(self, context):
        context["bar"] = True


@pytest.fixture(scope="function")
def templater():
    rv = Renoir()
    rv.use_extension(FooExtension)
    rv.use_extension(BarExtension)
    return rv


def test_config(templater):
    assert set(templater.loaders.keys()) == {".test"}
    assert len(templater.loaders[".test"]) == 1
    assert templater.loaders[".test"][0].__self__.__class__ == BarExtension

    assert len(templater.renderers) == 1
    assert templater.renderers[0].__self__.__class__ == FooExtension

    assert len(templater.contexts) == 1
    assert templater.contexts[0].__self__.__class__ == BarExtension

    assert set(templater.lexers.keys()) == {"foo"}
    assert isinstance(templater.lexers["foo"], FooLexer)


def test_load(templater):
    path, name = templater.preload("test.html.test")
    assert name == "test.html.test.bar"


def test_render(templater):
    source = templater.prerender("test", "<string>")
    assert source == "testfoo"


def test_context(templater):
    ctx = {"foo": False}
    templater.inject(ctx)
    assert ctx["foo"] is False
    assert ctx["bar"] is True


def test_lexers(templater):
    r = templater._render(source="{{foo asd}}")
    assert r == "asdasd"



# -*- coding: utf-8 -*-
"""
renoir.apis
-----------

Provides the main apis for the templating system.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

import os
import sys
from functools import reduce
from typing import Any, Dict, List, Optional, Tuple, Type

from .cache import TemplaterCache
from .constants import ESCAPES, MODES, NOFILEPATH
from .debug import make_traceback
from .errors import TemplateError, TemplateMissingError, TemplateSyntaxError
from .extensions import Extension
from .helpers import ParserCtx, TemplateReference, adict
from .parsing import HTMLIndentTemplateParser, HTMLTemplateParser, IndentTemplateParser, Lexer, TemplateParser
from .typing import ContextType, LoaderType, RenderType
from .writers import EscapeAllWriter, Writer


class Renoir:
    _writers = {ESCAPES.common: Writer, ESCAPES.all: EscapeAllWriter}

    def __init__(
        self,
        path: Optional[str] = None,
        loaders: Optional[Dict[str, List[LoaderType]]] = None,
        renderers: Optional[List[RenderType]] = None,
        contexts: Optional[List[ContextType]] = None,
        lexers: Optional[Dict[str, Lexer]] = None,
        delimiters: Tuple[str, str] = ("{{", "}}"),
        encoding: str = "utf8",
        mode: str = MODES.html,
        escape: str = ESCAPES.common,
        adjust_indent: bool = False,
        reload: bool = False,
        debug: bool = False,
    ):
        self.path = path or os.getcwd()
        self.loaders = loaders or {}
        self.renderers = renderers or []
        self.contexts = contexts or []
        self.lexers = lexers or {}
        self.delimiters = delimiters
        self.encoding = encoding
        self.mode = mode
        self.escape = escape
        self.indent = adjust_indent
        self.cache = TemplaterCache(self, reload=reload or debug)
        self._extensions = []
        self._extensions_env = {}
        self._configure()

    def _configure(self):
        self.writer_cls = self._writers.get(self.escape, self._writers[ESCAPES.common])
        if not self.indent:
            self.parser_cls = HTMLTemplateParser if self.mode == MODES.html else TemplateParser
        else:
            self.parser_cls = HTMLIndentTemplateParser if self.mode == MODES.html else IndentTemplateParser
        self.preload = self._preload if self.loaders else self._no_preload

    def __init_extension(self, ext_cls):
        namespace = ext_cls.namespace or ext_cls.__name__
        if namespace not in self._extensions_env:
            self._extensions_env[namespace] = adict()
        return namespace, self._extensions_env[namespace]

    def use_extension(self, ext_cls: Type[Extension], **config) -> Extension:
        if not issubclass(ext_cls, Extension):
            raise RuntimeError(f"{ext_cls.__name__} is an invalid Renoir extension")
        namespace, env = self.__init_extension(ext_cls)
        ext = ext_cls(self, namespace, env, config)
        if ext.file_extension:
            self.loaders[ext.file_extension] = self.loaders.get(ext.file_extension) or []
            self.loaders[ext.file_extension].append(ext.load)
        if ext._ext_render_:
            self.renderers.append(ext.render)
        if ext._ext_context_:
            self.contexts.append(ext.context)
        for name, lexer in ext.lexers.items():
            self.lexers[name] = lexer(ext=ext)
        self._extensions.append(ext)
        ext.on_load()
        self._configure()
        return ext

    def _preload(self, file_name, path=None):
        path = path or self.path
        file_extension = os.path.splitext(file_name)[1]
        return reduce(
            lambda args, loader: loader(args[0], args[1]), self.loaders.get(file_extension, []), (path, file_name)
        )

    def _no_preload(self, file_name, path=None):
        return (path or self.path, file_name)

    def _load(self, file_path):
        with open(file_path, "r", encoding=self.encoding) as file_obj:
            source = file_obj.read()
        return source

    def load(self, file_path):
        rv = self.cache.load.get(file_path)
        if not rv:
            try:
                rv = self._load(file_path)
            except Exception:
                raise TemplateMissingError(file_path)
            self.cache.load.set(file_path, rv)
        return rv

    def _prerender(self, source, name):
        return reduce(lambda source, renderer: renderer(source, name), self.renderers, source)

    def prerender(self, source, name):
        rv = self.cache.prerender.get(name, source)
        if not rv:
            rv = self._prerender(source, name)
            self.cache.prerender.set(name, source)
        return rv

    def parse(self, file_path, source, context):
        code, content = self.cache.parse.get(file_path, source)
        if not code:
            parser = self.parser_cls(
                self, source, name=file_path, scope=context, lexers=self.lexers, delimiters=self.delimiters
            )
            try:
                code = compile(parser.render(), os.path.split(file_path)[-1], "exec")
            except SyntaxError:
                parser_ctx = ParserCtx(file_path, parser.content)
                raise TemplateSyntaxError(parser_ctx, *sys.exc_info())
            content = parser.content
            self.cache.parse.set(file_path, source, code, content, parser.dependencies)
        return code, content

    def inject(self, context):
        for injector in self.contexts:
            injector(context)

    def _render(self, source="", file_path=NOFILEPATH, context=None):
        context = context or {}
        context["__writer__"] = self.writer_cls()
        try:
            code, content = self.parse(file_path, source, context)
        except (TemplateError, TemplateSyntaxError):
            make_traceback(sys.exc_info())
        self.inject(context)
        try:
            exec(code, context)
        except Exception:
            exc_info = sys.exc_info()
            try:
                parser_ctx = ParserCtx(file_path, content)
                template_ref = TemplateReference(parser_ctx, *exc_info)
            except Exception:
                template_ref = None
            context["__renoir_template__"] = template_ref
            make_traceback(exc_info)
        return context["__writer__"].body.getvalue()

    def render(self, template_file_name: str, context: Optional[Dict[str, Any]] = None) -> str:
        file_path = os.path.join(*self.preload(template_file_name))
        source = self.prerender(self.load(file_path), file_path)
        return self._render(source, file_path, context)



# -*- coding: utf-8 -*-
"""
renoir._internal
----------------

Provides internally used helpers and objects.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""


def reraise(tp, value, tb=None):
    if value.__traceback__ is not tb:
        raise value.with_traceback(tb)
    raise value



# -*- coding: utf-8 -*-
"""
renoir.cache
------------

Provides caching utils.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

import os

from ._shortcuts import hashlib_sha1


def make_hash(value):
    return hashlib_sha1(value).hexdigest()[:10]


class TemplaterCache:
    def __init__(self, templater, reload=False):
        self.templater = templater
        self.changes = reload
        self.load = LoaderCache(self)
        self.prerender = PrerenderCache(self)
        self.parse = ParserCache(self)


class InnerCache:
    def __init__(self, cache_interface):
        self.cache = cache_interface
        self.data = {}
        self._configure()

    def _configure(self):
        self.get = self.reloader_get if self.cache.changes else self.cached_get


class LoaderCache(InnerCache):
    def __init__(self, cache_interface):
        super().__init__(cache_interface)
        self.mtimes = {}

    def reloader_get(self, file_path):
        try:
            mtime = os.stat(file_path).st_mtime
        except Exception:
            return None
        old_time = self.mtimes.get(file_path, 0)
        if mtime > old_time:
            return None
        return self.cached_get(file_path)

    def cached_get(self, file_path):
        return self.data.get(file_path)

    def set(self, file_path, source):
        self.data[file_path] = source
        self.mtimes[file_path] = os.stat(file_path).st_mtime


class HashableCache(InnerCache):
    def __init__(self, cache_interface):
        super().__init__(cache_interface)
        self.hashes = {}

    def reloader_get(self, name, source):
        hashed = make_hash(source)
        if self.hashes.get(name) != hashed:
            return None
        return self.cached_get(name, source)

    def cached_get(self, name, source):
        return self.data.get(name)

    def set(self, name, source):
        self.data[name] = source
        if self.cache.changes:
            self.hashes[name] = make_hash(source)


class PrerenderCache(HashableCache):
    pass


class ParserCache(HashableCache):
    def __init__(self, cache_interface):
        super().__init__(cache_interface)
        self.cdata = {}
        self.dependencies = {}

    def _expired_dependency(self, name, preload_params):
        path, file_name = self.cache.templater.preload(name, **preload_params)
        file_path = os.path.join(path, file_name)
        if os.stat(file_path).st_mtime != self.cache.load.mtimes[file_path]:
            return True
        return False

    def reloader_get(self, name, source):
        hashed = make_hash(source)
        if self.hashes.get(name) != hashed:
            return None, None
        for dep_name, dep_preload_params in self.dependencies[name].values():
            if self._expired_dependency(dep_name, dep_preload_params):
                return None, None
        return self.cached_get(name, source)

    def cached_get(self, name, source):
        return self.data.get(name), self.cdata.get(name)

    def set(self, name, source, compiled, content, dependencies):
        self.data[name] = compiled
        self.cdata[name] = content
        if self.cache.changes:
            self.hashes[name] = make_hash(source)
            self.dependencies[name] = dependencies



# -*- coding: utf-8 -*-
"""
renoir.extensions
-----------------

Provides base classes to create extensions.

:copyright: 2014 Giovanni Barillari
:license: BSD-3-Clause
"""

from typing import Any, Dict, Optional, Tuple, Type

from .parsing.lexers import Lexer


class MetaExtension(type):
    _ext_methods_ = {"load", "render", "context"}

    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        if not bases:
            return new_class
        declared_methods = cls._ext_methods_ & set(attrs.keys())
        new_class._ext_declared_methods_ = declared_methods
        all_methods = set()
        for base in reversed(new_class.__mro__[:-2]):
            if hasattr(base, "_ext_declared_methods_"):
                all_methods = all_methods | base._ext_declared_methods_
        all_methods = all_methods | declared_methods
        new_class._ext_all_methods_ = all_methods
        new_class._ext_render_ = "render" in all_methods
        new_class._ext_context_ = "context" in all_methods
        return new_class


class Extension(metaclass=MetaExtension):
    namespace: Optional[str] = None
    file_extension: Optional[str] = None
    lexers: Dict[str, Type[Lexer]] = {}
    default_config: Dict[str, Any] = {}

    def __init__(self, templater, namespace, env, config=None):
        self.templater = templater
        self.namespace = namespace
        self.env = env
        self.config = config or {}
        if isinstance(self.file_extension, str) and not self.file_extension.startswith("."):
            self.file_extension = "." + self.file_extension
        self.__load_config()

    def __load_config(self):
        for key, default_val in self.default_config.items():
            self.config[key] = self.config.get(key, default_val)

    def on_load(self):
        pass

    def load(self, path: str, file_name: str) -> Tuple[str, str]:
        return path, file_name

    def render(self, source: str, name: str) -> str:
        return source

    def context(self, context: Dict[str, Any]):
        pass
